Package: combss
Type: Package
Title: Continuous Optimisation Towards Best Subset Selection
Version: 0.1.0
Authors@R: c(person(given = "Benoit", family = "Liquet", email = "benoit.liquet@univ-pau.fr", role = c("aut", "cre"), comment = c(ORCID = "0000-0002-8136-2294")),person(given = "Anant", family = "Mathur", role = "aut"), person(given = "Sarat", family = "Moka", role = "aut"))
Author: Benoit Liquet [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-8136-2294>),
  Anant Mathur [aut],
  Sarat Moka [aut]
Maintainer: Benoit Liquet <benoit.liquet@univ-pau.fr>
Description: Best subset selection in generalised linear models via continuous
    optimisation. Reformulates the NP-hard discrete subset selection problem as
    a continuous optimisation over the hypercube [0,1]^p, solved via a
    Frank-Wolfe homotopy algorithm with closed-form ridge inner solves.
    Supports linear (Gaussian), binary logistic, and multinomial regression.
    For methodological details see Moka, Liquet, Zhu and Muller (2024)
    <doi:10.1007/s11222-024-10387-8> and Mathur, Liquet, Muller and Moka
    (2026) <doi:10.48550/arXiv.2603.21952>.
License: GPL-3
URL: https://github.com/benoit-liquet/combss
Encoding: UTF-8
Imports: glmnet (>= 4.0), stats
Suggests: testthat (>= 3.0.0), knitr, rmarkdown
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-06 06:20:54 UTC; benoit
Repository: CRAN
Date/Publication: 2026-05-11 19:10:26 UTC
Built: R 4.5.2; ; 2026-05-11 21:57:42 UTC; unix

Package: combss
Type: Package
Title: Continuous Optimisation Towards Best Subset Selection
Version: 0.1.0
Authors@R: c(person(given = "Benoit", family = "Liquet", email = "benoit.liquet@univ-pau.fr", role = c("aut", "cre"), comment = c(ORCID = "0000-0002-8136-2294")),person(given = "Anant", family = "Mathur", role = "aut"), person(given = "Sarat", family = "Moka", role = "aut"))
Author: Benoit Liquet [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-8136-2294>),
  Anant Mathur [aut],
  Sarat Moka [aut]
Maintainer: Benoit Liquet <benoit.liquet@univ-pau.fr>
Description: Best subset selection in generalised linear models via continuous
    optimisation. Reformulates the NP-hard discrete subset selection problem as
    a continuous optimisation over the hypercube [0,1]^p, solved via a
    Frank-Wolfe homotopy algorithm with closed-form ridge inner solves.
    Supports linear (Gaussian), binary logistic, and multinomial regression.
    For methodological details see Moka, Liquet, Zhu and Muller (2024)
    <doi:10.1007/s11222-024-10387-8> and Mathur, Liquet, Muller and Moka
    (2026) <doi:10.48550/arXiv.2603.21952>.
License: GPL-3
URL: https://github.com/benoit-liquet/combss
Encoding: UTF-8
Imports: glmnet (>= 4.0), stats
Suggests: testthat (>= 3.0.0), knitr, rmarkdown
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-06 06:20:54 UTC; benoit
Repository: CRAN
Date/Publication: 2026-05-11 19:10:26 UTC
Built: R 4.6.0; ; 2026-05-11 20:32:07 UTC; unix

S3method(coef,combss)
S3method(predict,combss)
S3method(print,combss)
S3method(summary,combss)
export(combss)
export(combss_cv)
importFrom(stats,aggregate)
importFrom(stats,coef)
importFrom(stats,rnorm)

#' combss: Best Subset Selection for Generalised Linear Models via Continuous Optimisation
#'
#' Best subset selection in linear, binary logistic, and multinomial logistic
#' regression via a Frank-Wolfe homotopy algorithm. The algorithm reformulates
#' the NP-hard discrete subset selection problem as a continuous optimisation
#' over the hypercube `[0, 1]^p`, scaling to high-dimensional settings with
#' p >> n. Inner ridge solves use \pkg{glmnet}.
#'
#' @keywords internal
#' @importFrom stats aggregate coef rnorm
"_PACKAGE"

## usethis namespace: start
## usethis namespace: end
NULL

#' Best subset selection for generalised linear models via continuous
#' optimisation (COMBSS).
#' Reformulates the NP-hard discrete subset selection problem as a continuous
#' optimisation over the hypercube `[0, 1]^p`, solved via a Frank-Wolfe
#' homotopy algorithm. The inner ridge problem is solved with `glmnet`. For
#' each subset size `k = 1, ..., q`, COMBSS returns the selected feature set;
#' if validation data is supplied, the best `k` is chosen by validation MSE
#' (Gaussian) or accuracy (binomial / multinomial).
#'
#' @param x Numeric design matrix `(n, p)` without an intercept column.
#' @param y Response. Numeric for `family = "gaussian"`; `{0, 1}` (or two-level
#'   factor) for `family = "binomial"`; factor or integer in `{1, ..., C}` for
#'   `family = "multinomial"`.
#' @param family One of `"gaussian"` (alias: `"linear"`), `"binomial"`,
#'   `"multinomial"`.
#' @param q Maximum subset size. Defaults to `min(n, p)`.
#' @param x_val,y_val Optional validation data. When provided, the best `k`
#'   is selected by validation metric and `coef_` is returned for the best
#'   subset.
#' @param lam_ridge Ridge penalty for the inner solver. Default `0`.
#' @param Niter Number of homotopy iterations. Default `25`.
#' @param alpha Frank-Wolfe step size. Default `0.01`.
#' @param scale Column-normalise `x` before the algorithm. Default `TRUE`.
#' @param mandatory Optional integer vector of 1-indexed columns of `x` that
#'   must be included in every model.
#' @param verbose Print progress. Default `FALSE`.
#'
#' @return An object of class `"combss"`, a list with elements:
#'   - `family`: one of `"gaussian"`, `"binomial"`, `"multinomial"`
#'   - `subset_list`: list of length up to `q`; element `k` is an integer
#'     vector of selected column indices (1-indexed) for subset size `k`
#'   - `subset`: best subset (requires validation data; otherwise `NULL`)
#'   - `mse` / `accuracy`: validation metric for the best subset
#'     (requires validation data)
#'   - `lam_ridge`: ridge penalty used
#'   - `q`: maximum subset size evaluated
#'
#' @references
#' Moka, S., Liquet, B., Zhu, H. and Muller, S. (2024).
#' COMBSS: best subset selection via continuous optimization.
#' *Statistics and Computing*. \doi{10.1007/s11222-024-10387-8}
#'
#' Mathur, A., Liquet, B., Muller, S. and Moka, S. (2026).
#' Parsimonious Subset Selection for Generalized Linear Models with
#' Biomedical Applications. \href{https://arxiv.org/abs/2603.21952}{arXiv:2603.21952}.
#'
#' @examples
#' set.seed(1)
#' n <- 200; p <- 30
#' beta <- c(3, 2, 1.5, 1, 0.5, rep(0, p - 5))
#' x <- matrix(rnorm(n * p), n, p)
#' y <- as.numeric(x %*% beta + rnorm(n) * 0.5)
#' fit <- combss(x, y, family = "gaussian", q = 10)
#' fit$subset_list[1:5]
#'
#' @export
combss <- function(x, y,
                   family = c("gaussian", "binomial", "multinomial"),
                   q = NULL,
                   x_val = NULL, y_val = NULL,
                   lam_ridge = 0,
                   Niter = 25,
                   alpha = 0.01,
                   scale = TRUE,
                   mandatory = NULL,
                   verbose = FALSE) {

  family <- resolve_family(family)
  x <- as.matrix(x)
  n <- nrow(x); p <- ncol(x)
  if (is.null(q)) q <- min(n, p)

  if (family == "gaussian") {
    y <- as.numeric(y)
    solve_inner <- solve_inner_gaussian
  } else if (family == "binomial") {
    y <- as.numeric(y)
    solve_inner <- solve_inner_binomial
  } else {
    y <- as.factor(y)
    solve_inner <- solve_inner_multinomial
  }

  if (is.null(mandatory)) {
    mand_idx <- integer(0)
    sel_idx  <- seq_len(p)
  } else {
    mand_idx <- sort(as.integer(mandatory))
    sel_idx  <- setdiff(seq_len(p), mand_idx)
  }
  p_sel <- length(sel_idx)
  if (q > p_sel + length(mand_idx)) q <- p_sel + length(mand_idx)

  X_run <- if (scale) scale_design_matrix(x)$X_scaled else x
  cal <- calibrate_penalty_schedule(X_run[, sel_idx, drop = FALSE], n, family)

  if (verbose) {
    message(sprintf("Calibration: nu_max=%.4f | delta_conc=%.6f | delta_min=%.8f | delta_max=%.6f",
                    cal$nu_max, cal$delta_conc, cal$delta_min, cal$delta_max))
  }

  subset_list <- vector("list", q)
  for (k in seq_len(q)) {
    if (k <= length(mand_idx)) {
      subset_list[[k]] <- sort(mand_idx[seq_len(k)])
      next
    }
    k_relax <- k - length(mand_idx)
    t_init <- rep(k_relax / p_sel, p_sel)
    res <- fw_loop(t_init, k_relax, X_run, y,
                   cal$delta_min, cal$delta_max,
                   Niter, alpha, lam_ridge,
                   mand_idx, sel_idx, solve_inner)
    subset_list[[k]] <- res$selected
    if (verbose) {
      message(sprintf("  k = %2d  selected: %s",
                      k, paste(res$selected, collapse = ",")))
    }
  }

  out <- list(
    family       = family,
    subset_list  = subset_list,
    subset       = NULL,
    lam_ridge    = lam_ridge,
    q            = q,
    n            = n,
    p            = p,
    feature_names = colnames(x)
  )
  out[[if (family == "gaussian") "mse" else "accuracy"]] <- NULL

  if (!is.null(x_val) && !is.null(y_val)) {
    x_val <- as.matrix(x_val)
    y_val_raw <- y_val
    metric <- numeric(q)
    coefs  <- vector("list", q)
    for (k in seq_len(q)) {
      idx <- subset_list[[k]]
      pred <- refit_predict(x, y, x_val, idx, family)
      if (family == "gaussian") {
        metric[k] <- mean((as.numeric(y_val_raw) - pred)^2)
      } else if (family == "binomial") {
        metric[k] <- mean(pred == as.numeric(y_val_raw))
      } else {
        metric[k] <- mean(pred == as.character(y_val_raw))
      }
    }
    best_k <- if (family == "gaussian") which.min(metric) else which.max(metric)
    out$subset <- subset_list[[best_k]]
    if (family == "gaussian") {
      out$mse <- metric[best_k]
      out$mse_path <- metric
    } else {
      out$accuracy <- metric[best_k]
      out$accuracy_path <- metric
    }
    out$best_k <- best_k
  }

  class(out) <- "combss"
  out
}

make_linear_data <- function(n = 200, p = 30, k_true = 5, seed = 1) {
  set.seed(seed)
  beta <- c(c(3, 2, 1.5, 1, 0.5)[seq_len(k_true)], rep(0, p - k_true))
  x <- matrix(rnorm(n * p), n, p)
  y <- as.numeric(x %*% beta + rnorm(n) * 0.5)
  list(x = x, y = y, beta = beta, n = n, p = p)
}

test_that("gaussian FW recovers the active features at k = k_true", {
  d <- make_linear_data()
  fit <- combss(d$x, d$y, family = "gaussian", q = 10, verbose = FALSE)
  expect_s3_class(fit, "combss")
  expect_length(fit$subset_list, 10)
  expect_setequal(fit$subset_list[[5]], 1:5)
})

test_that("'linear' is an alias for 'gaussian'", {
  d <- make_linear_data()
  f1 <- combss(d$x, d$y, family = "gaussian", q = 5, verbose = FALSE)
  f2 <- combss(d$x, d$y, family = "linear",   q = 5, verbose = FALSE)
  expect_identical(f1$subset_list, f2$subset_list)
  expect_equal(f1$family, "gaussian")
  expect_equal(f2$family, "gaussian")
})

test_that("validation data selects the best k and stores the metric", {
  d <- make_linear_data()
  itr <- 1:140; iva <- 141:200
  fit <- combss(d$x[itr, ], d$y[itr],
                x_val = d$x[iva, ], y_val = d$y[iva],
                family = "gaussian", q = 10, verbose = FALSE)
  expect_false(is.null(fit$subset))
  expect_false(is.null(fit$mse))
  expect_length(fit$mse_path, 10)
  expect_equal(which.min(fit$mse_path), fit$best_k)
})

test_that("binomial FW separates labels in a clean simulation", {
  d <- make_linear_data()
  ybin <- as.numeric(plogis(d$x %*% d$beta) > 0.5)
  fit <- combss(d$x, ybin, family = "binomial", q = 8, verbose = FALSE)
  expect_s3_class(fit, "combss")
  expect_setequal(intersect(fit$subset_list[[5]], 1:5), 1:5)
})

test_that("multinomial FW recovers active features", {
  d <- make_linear_data()
  ymulti <- cut(as.numeric(d$x %*% d$beta),
                breaks = c(-Inf, -1, 1, Inf),
                labels = c("a", "b", "c"))
  fit <- combss(d$x, ymulti, family = "multinomial", q = 8, verbose = FALSE)
  expect_s3_class(fit, "combss")
  expect_true(length(fit$subset_list[[8]]) == 8)
})

test_that("coef() returns the chosen subset", {
  d <- make_linear_data()
  fit <- combss(d$x, d$y, family = "gaussian", q = 5, verbose = FALSE)
  expect_setequal(coef(fit, k = 3), 1:3)
})

test_that("invalid family is rejected", {
  d <- make_linear_data()
  expect_error(combss(d$x, d$y, family = "poisson", q = 3),
               "must be one of")
})

#' Inner ridge solver for binary logistic regression.
#'
#' Solves the t-parameterised inner problem via `glmnet` with `alpha = 0`.
#' Returns the coefficient vector (including intercept) at the optimum.
#'
#' @param t Numeric vector in (0, 1)^p_sel: relaxed inclusion vector for the
#'   selectable columns.
#' @param X Numeric matrix (no intercept column).
#' @param y Binary response in {0, 1}.
#' @param delta Current homotopy concentration parameter.
#' @param lambda Ridge penalty.
#' @param mand_idx Integer vector of mandatory column indices (1-based).
#' @param sel_idx Integer vector of selectable column indices (1-based).
#' @return Numeric vector: `(beta_0, beta_1, ..., beta_p)`.
#' @keywords internal
#' @noRd
solve_inner_binomial <- function(t, X, y, delta, lambda,
                                 mand_idx = integer(0),
                                 sel_idx = seq_len(ncol(X))) {
  p <- ncol(X)
  omega <- rep(0, p)
  if (length(mand_idx) > 0) {
    omega[mand_idx] <- lambda
  }
  omega[sel_idx] <- (lambda + delta) / (t^2) - delta
  omega <- pmax(omega, 1e-12)

  omega_sum <- sum(omega)
  if (omega_sum < 1e-12) omega_sum <- 1e-12
  penalty_factors <- p * omega / omega_sum
  lambda_glmnet <- 2 * omega_sum / p

  fit <- glmnet::glmnet(
    x = X, y = y, family = "binomial", alpha = 0,
    lambda = lambda_glmnet, penalty.factor = penalty_factors,
    intercept = TRUE, standardize = FALSE,
    thresh = 1e-12, maxit = 1e6
  )
  as.numeric(coef(fit))
}

#' Inner ridge solver for multinomial logistic regression.
#'
#' Uses `glmnet` with `family = "multinomial"` and grouped penalisation so
#' the same sparsity pattern is shared across classes.
#'
#' @inheritParams solve_inner_binomial
#' @param y Factor or integer-coded response.
#' @return Numeric matrix `(C, p + 1)` of class-by-coefficient values
#'   (column 1 is the intercept).
#' @keywords internal
#' @noRd
solve_inner_multinomial <- function(t, X, y, delta, lambda,
                                    mand_idx = integer(0),
                                    sel_idx = seq_len(ncol(X))) {
  p <- ncol(X)
  omega <- rep(0, p)
  if (length(mand_idx) > 0) {
    omega[mand_idx] <- lambda
  }
  omega[sel_idx] <- (lambda + delta) / (t^2) - delta
  omega <- pmax(omega, 1e-12)

  omega_sum <- sum(omega)
  if (omega_sum < 1e-12) omega_sum <- 1e-12
  penalty_factors <- p * omega / omega_sum
  lambda_glmnet <- 2 * omega_sum / p

  fit <- glmnet::glmnet(
    x = X, y = y, family = "multinomial",
    type.multinomial = "grouped", alpha = 0,
    lambda = lambda_glmnet, penalty.factor = penalty_factors,
    intercept = TRUE, standardize = FALSE,
    thresh = 1e-10, maxit = 1e5
  )
  cf <- coef(fit)
  do.call(rbind, lapply(cf, function(m) as.numeric(m)))
}

#' Inner ridge solver for linear (Gaussian) regression.
#'
#' Uses `glmnet` with `family = "gaussian"` and `alpha = 0`. Returns the
#' coefficient vector (including intercept) at the optimum.
#'
#' @inheritParams solve_inner_binomial
#' @param y Numeric response.
#' @return Numeric vector: `(beta_0, beta_1, ..., beta_p)`.
#' @keywords internal
#' @noRd
solve_inner_gaussian <- function(t, X, y, delta, lambda,
                                 mand_idx = integer(0),
                                 sel_idx = seq_len(ncol(X))) {
  p <- ncol(X)
  omega <- rep(0, p)
  if (length(mand_idx) > 0) {
    omega[mand_idx] <- lambda
  }
  omega[sel_idx] <- (lambda + delta) / (t^2) - delta
  omega <- pmax(omega, 1e-12)

  omega_sum <- sum(omega)
  if (omega_sum < 1e-12) omega_sum <- 1e-12
  penalty_factors <- p * omega / omega_sum
  lambda_glmnet <- 2 * omega_sum / p

  fit <- glmnet::glmnet(
    x = X, y = y, family = "gaussian", alpha = 0,
    lambda = lambda_glmnet, penalty.factor = penalty_factors,
    intercept = TRUE, standardize = FALSE,
    thresh = 1e-12, maxit = 1e6
  )
  as.numeric(coef(fit))
}

#' Danskin envelope gradient.
#'
#' Computes `df/dt_j = -2 (lambda + delta) * sum_c xi_{j,c}^2 / t_j^3`
#' where the sum collapses to a single term in the binomial / Gaussian case.
#'
#' @param xi Numeric vector (binomial / Gaussian) or matrix (multinomial)
#'   of inner coefficients including intercept(s).
#' @param t Numeric vector of relaxed inclusion values.
#' @param delta,lambda Current homotopy and ridge parameters.
#' @param sel_idx Selectable column indices in the original `X`.
#' @return Gradient vector of length `length(t)`.
#' @keywords internal
#' @noRd
grad_envelope <- function(xi, t, delta, lambda, sel_idx) {
  if (is.matrix(xi)) {
    # multinomial: column 1 is intercept; selectable columns are sel_idx + 1
    xi_sel <- xi[, sel_idx + 1, drop = FALSE]
    xi_sq_sum <- colSums(xi_sel^2)
  } else {
    xi_sel <- xi[sel_idx + 1]
    xi_sq_sum <- xi_sel^2
  }
  -2 * (lambda + delta) * xi_sq_sum / (t^3)
}

#' Column-wise L2 normalisation of a design matrix.
#'
#' Scales each column of `X` to unit L2 norm. Columns of zero norm are left
#' unchanged. Returned alongside the scaling vector so coefficients can be
#' rescaled if needed.
#'
#' @param X Numeric matrix.
#' @return List with `X_scaled` (the scaled matrix) and `v` (the original
#'   per-column L2 norms).
#' @keywords internal
#' @noRd
scale_design_matrix <- function(X) {
  v <- sqrt(colSums(X^2))
  v[v == 0] <- 1
  X_scaled <- sweep(X, 2, v, "/")
  list(X_scaled = X_scaled, v = v)
}

#' Power iteration for the largest eigenvalue of `t(X) %*% X`.
#'
#' Used to calibrate the Frank-Wolfe penalty schedule.
#'
#' @param X Numeric matrix.
#' @param n_iter Maximum number of power iterations.
#' @param tol Relative tolerance.
#' @return Estimate of `nu_max`.
#' @keywords internal
#' @noRd
compute_nu_max <- function(X, n_iter = 200, tol = 1e-8,seed = NULL) {
  if(!is.null(seed)){
    set.seed(seed)
  }
  p <- ncol(X)
  v <- rnorm(p)
  v <- v / sqrt(sum(v^2))
  nu_old <- 0
  for (i in seq_len(n_iter)) {
    Xv   <- X %*% v
    XtXv <- crossprod(X, Xv)
    nu   <- sum(Xv^2)
    v    <- as.numeric(XtXv)
    v    <- v / sqrt(sum(v^2))
    if (abs(nu - nu_old) < tol * max(nu, 1)) break
    nu_old <- nu
  }
  nu
}

#' Calibrate the Frank-Wolfe homotopy delta schedule.
#'
#' Computes `delta_min` and `delta_max` from the spectral norm of `X` and
#' the family-specific concentration constant.
#'
#' @param X_u Normalised design matrix (only the relaxed columns).
#' @param n Number of training rows.
#' @param family One of `"gaussian"`, `"binomial"`, `"multinomial"`.
#' @return List with `nu_max`, `delta_conc`, `delta_max`, `delta_min`.
#' @keywords internal
#' @noRd
calibrate_penalty_schedule <- function(X_u, n, family) {
  nu_max     <- compute_nu_max(X_u)
  delta_conc <- switch(family,
                       gaussian    = 3 * nu_max / (8 * n),
                       binomial    = nu_max / (8 * n),
                       multinomial = nu_max / (4 * n),
                       stop("Unknown family: ", family))
  delta_max  <- delta_conc
  delta_min  <- 1e-3 * delta_conc
  list(nu_max = nu_max, delta_conc = delta_conc,
       delta_max = delta_max, delta_min = delta_min)
}

#' Resolve the family argument, accepting `"linear"` as an alias for `"gaussian"`.
#'
#' @param family User-supplied family string.
#' @return One of `"gaussian"`, `"binomial"`, `"multinomial"`.
#' @keywords internal
#' @noRd
resolve_family <- function(family) {
  family <- tolower(family[1])
  if (family == "linear") family <- "gaussian"
  if (!family %in% c("gaussian", "binomial", "multinomial")) {
    stop("`family` must be one of \"gaussian\" (or \"linear\"), \"binomial\", \"multinomial\".",
         call. = FALSE)
  }
  family
}

#' @export
print.combss <- function(x, ...) {
  cat("COMBSS fit\n")
  cat("  family:    ", x$family, "\n", sep = "")
  cat("  n, p:      ", x$n, ", ", x$p, "\n", sep = "")
  cat("  q:         ", x$q, "\n", sep = "")
  cat("  lam_ridge: ", x$lam_ridge, "\n", sep = "")
  if (!is.null(x$subset)) {
    cat("  best k:    ", length(x$subset), "\n", sep = "")
    if (!is.null(x$mse))
      cat("  mse:       ", signif(x$mse, 5), "\n", sep = "")
    if (!is.null(x$accuracy))
      cat("  accuracy:  ", signif(x$accuracy, 5), "\n", sep = "")
    cat("  subset:    ", paste(x$subset, collapse = ", "), "\n", sep = "")
  } else {
    cat("  (no validation data; subset_list only)\n")
  }
  invisible(x)
}

#' @export
summary.combss <- function(object, ...) {
  print(object)
  cat("\nSubset path:\n")
  for (k in seq_along(object$subset_list)) {
    feats <- object$subset_list[[k]]
    line <- sprintf("  k=%2d  features: %s", k, paste(feats, collapse = ","))
    if (!is.null(object$mse_path))
      line <- paste0(line, sprintf("  mse=%.4f", object$mse_path[k]))
    if (!is.null(object$accuracy_path))
      line <- paste0(line, sprintf("  acc=%.4f", object$accuracy_path[k]))
    cat(line, "\n")
  }
  invisible(object)
}

#' Extract the subset of selected feature indices.
#'
#' @param object A `combss` object.
#' @param k Subset size to return. Defaults to the best `k` if validation
#'   data was used, else the largest `k` evaluated.
#' @param ... Unused.
#' @return Integer vector of selected column indices (1-indexed).
#' @examples
#' set.seed(1)
#' x <- matrix(rnorm(100 * 20), 100, 20)
#' y <- as.numeric(x[, 1] + x[, 2] + rnorm(100) * 0.1)
#' fit <- combss(x, y, family = "gaussian", q = 5)
#' coef(fit, k = 2)
#' @export
coef.combss <- function(object, k = NULL, ...) {
  if (is.null(k)) {
    k <- if (!is.null(object$best_k)) object$best_k else object$q
  }
  object$subset_list[[k]]
}

#' Predict from a `combss` fit.
#'
#' Refits the chosen subset on the original training data via OLS or ridge
#' (Gaussian / binomial / multinomial as appropriate) and predicts on `newx`.
#' Note: the training data is not stored on the fit object, so `newx` must be
#' supplied along with `x_train` and `y_train`.
#'
#' @param object A `combss` object.
#' @param newx Numeric matrix of new observations.
#' @param x_train,y_train Original training data used to refit on the chosen
#'   subset.
#' @param k Subset size to use. Defaults to the best `k` if validation data
#'   was used, else the largest `k` evaluated.
#' @param ... Unused.
#' @return Numeric vector of predictions for `newx`.
#' @examples
#' set.seed(1)
#' x <- matrix(rnorm(100 * 20), 100, 20)
#' y <- as.numeric(x[, 1] + x[, 2] + rnorm(100) * 0.1)
#' fit <- combss(x, y, family = "gaussian", q = 5)
#' newx <- matrix(rnorm(5 * 20), 5, 20)
#' predict(fit, newx = newx, x_train = x, y_train = y, k = 2)
#' @export
predict.combss <- function(object, newx, x_train, y_train, k = NULL, ...) {
  if (missing(x_train) || missing(y_train))
    stop("`x_train` and `y_train` are required to refit on the chosen subset.",
         call. = FALSE)
  if (is.null(k)) {
    k <- if (!is.null(object$best_k)) object$best_k else object$q
  }
  idx <- object$subset_list[[k]]
  refit_predict(as.matrix(x_train), y_train, as.matrix(newx),
                idx, object$family)
}

#' Select the ridge penalty for COMBSS by leave-one-out cross-validation.
#'
#' Mirrors `combss.cv.select_lambda` in the Python `combss` package. For each
#' candidate `lambda` and each subset size `k = 1, ..., q`, runs COMBSS to
#' obtain a subset and evaluates LOOCV error of a refit. Returns the best
#' `lambda` overall and the best `lambda` per `k`.
#'
#' @param x Numeric design matrix `(n, p)`.
#' @param y Response (numeric / binary / factor depending on `family`).
#' @param family One of `"gaussian"` (alias `"linear"`), `"binomial"`,
#'   `"multinomial"`.
#' @param q Maximum subset size.
#' @param lambda_grid Candidate ridge penalties. Defaults to
#'   `c(0, 10^seq(-3, 1, length.out = 10))`.
#' @param Niter Number of homotopy iterations passed to `combss()`.
#' @param verbose Print progress.
#'
#' @return List with:
#'   - `best_lambda`: ridge penalty with best mean LOOCV metric across `k`
#'   - `best_lambda_per_k`: numeric vector, length `q`
#'   - `results`: data frame `(lambda, k, loocv_metric)`
#'
#' @export
combss_cv <- function(x, y,
                      family = c("gaussian", "binomial", "multinomial"),
                      q,
                      lambda_grid = NULL,
                      Niter = 25,
                      verbose = FALSE) {
  family <- resolve_family(family)
  if (is.null(lambda_grid)) lambda_grid <- c(0, 10^seq(-3, 1, length.out = 10))

  x <- as.matrix(x)
  n <- nrow(x)
  results <- list()
  best_lambda_per_k <- rep(NA_real_, q)
  best_metric_per_k <- rep(NA_real_, q)

  for (lam in lambda_grid) {
    if (verbose) message(sprintf("lambda = %g", lam))
    fit <- combss(x, y, family = family, q = q,
                  lam_ridge = lam, Niter = Niter, verbose = FALSE)
    for (k in seq_len(q)) {
      idx <- fit$subset_list[[k]]
      metric_k <- loocv_metric(x, y, idx, family)
      results[[length(results) + 1]] <- data.frame(
        lambda = lam, k = k, loocv_metric = metric_k
      )
      better <- if (family == "gaussian")
        is.na(best_metric_per_k[k]) || metric_k < best_metric_per_k[k]
      else
        is.na(best_metric_per_k[k]) || metric_k > best_metric_per_k[k]
      if (better) {
        best_lambda_per_k[k] <- lam
        best_metric_per_k[k] <- metric_k
      }
    }
  }

  results <- do.call(rbind, results)
  agg <- aggregate(results$loocv_metric, by = list(lambda = results$lambda),
                   FUN = if (family == "gaussian") mean else mean)
  best_lambda <- if (family == "gaussian")
    agg$lambda[which.min(agg$x)] else agg$lambda[which.max(agg$x)]

  list(best_lambda = best_lambda,
       best_lambda_per_k = best_lambda_per_k,
       results = results)
}

#' LOOCV metric for a fixed subset.
#'
#' Gaussian: hat-matrix shortcut for ridge LOOCV MSE.
#' Binomial / multinomial: explicit leave-one-out refit (slow but simple).
#'
#' @keywords internal
#' @noRd
loocv_metric <- function(x, y, idx, family) {
  n <- nrow(x)
  if (length(idx) == 0) {
    if (family == "gaussian") return(mean((y - mean(y))^2))
    return(mean(as.character(y) == names(which.max(table(y)))))
  }

  X_sub <- cbind(1, x[, idx, drop = FALSE])

  if (family == "gaussian") {
    XtX <- crossprod(X_sub)
    H <- X_sub %*% solve(XtX + diag(1e-8, ncol(XtX))) %*% t(X_sub)
    yhat <- as.numeric(H %*% as.numeric(y))
    h_diag <- pmin(diag(H), 1 - 1e-8)
    err <- (as.numeric(y) - yhat) / (1 - h_diag)
    return(mean(err^2))
  }

  acc <- 0
  for (i in seq_len(n)) {
    pred_i <- refit_predict(x[-i, , drop = FALSE], y[-i],
                            x[i, , drop = FALSE], idx, family)
    if (family == "binomial") {
      acc <- acc + (pred_i == as.numeric(y[i]))
    } else {
      acc <- acc + (pred_i == as.character(y[i]))
    }
  }
  acc / n
}

#' Frank-Wolfe homotopy loop for a single subset size `k`.
#'
#' This is an internal helper shared by all three GLM families.
#'
#' @param t_init Initial relaxed inclusion vector of length `p_sel`.
#' @param k Target subset size.
#' @param X Normalised design matrix (no intercept column).
#' @param y Response.
#' @param delta_min,delta_max Homotopy schedule endpoints.
#' @param Niter Number of homotopy iterations (loop runs `2 * Niter` steps).
#' @param alpha Frank-Wolfe step size.
#' @param lambda Ridge penalty.
#' @param mand_idx,sel_idx Mandatory and selectable column indices.
#' @param solve_inner Inner ridge solver function.
#' @return List with `selected` (1-indexed columns of `X`), `t`, `s`.
#' @keywords internal
#' @noRd
fw_loop <- function(t_init, k, X, y,
                    delta_min, delta_max,
                    Niter, alpha, lambda,
                    mand_idx, sel_idx,
                    solve_inner) {
  p_sel <- length(sel_idx)
  r <- (delta_max / delta_min)^(1 / Niter)
  t <- t_init
  s <- rep(0, p_sel)

  for (i in seq_len(2 * Niter)) {
    delta <- min(delta_min * r^i, delta_max)
    xi <- solve_inner(t, X, y, delta, lambda, mand_idx, sel_idx)
    g  <- grad_envelope(xi, t, delta, lambda, sel_idx)
    s <- rep(0, p_sel)
    s[order(g)[seq_len(k)]] <- 1
    t <- (1 - alpha) * t + alpha * s
    t <- pmin(pmax(t, 1e-4), 1 - 1e-4)
  }

  selected_sel  <- which(s == 1)
  selected_full <- sel_idx[selected_sel]
  if (length(mand_idx) > 0) {
    selected_full <- sort(c(mand_idx, selected_full))
  }
  list(selected = selected_full, t = t, s = s)
}

#' Refit on a subset of features and predict on test data.
#'
#' @keywords internal
#' @noRd
refit_predict <- function(X_train, y_train, X_test, idx, family) {
  if (length(idx) == 0) {
    if (family == "gaussian") {
      pred <- rep(mean(y_train), nrow(X_test))
    } else {
      tab <- table(y_train)
      majority <- names(which.max(tab))
      pred <- rep(if (family == "binomial") as.numeric(majority) else majority,
                  nrow(X_test))
    }
    return(pred)
  }

  X_sub_train <- X_train[, idx, drop = FALSE]
  X_sub_test  <- X_test[, idx, drop = FALSE]

  if (family == "gaussian") {
    if (ncol(X_sub_train) == 1) {
      df <- data.frame(y = y_train, x = as.numeric(X_sub_train))
      fit <- stats::lm(y ~ x, data = df)
      return(as.numeric(stats::predict(fit,
        newdata = data.frame(x = as.numeric(X_sub_test)))))
    }
    fit <- glmnet::glmnet(x = X_sub_train, y = y_train, family = "gaussian",
                          alpha = 0, lambda = 1e-8,
                          intercept = TRUE, standardize = FALSE)
    return(as.numeric(stats::predict(fit, newx = X_sub_test, s = 1e-8)))
  }

  if (family == "binomial") {
    if (ncol(X_sub_train) == 1) {
      df <- data.frame(y = y_train, x = as.numeric(X_sub_train))
      fit <- stats::glm(y ~ x, data = df, family = stats::binomial())
      prob <- stats::predict(fit,
        newdata = data.frame(x = as.numeric(X_sub_test)),
        type = "response")
    } else {
      fit <- glmnet::glmnet(x = X_sub_train, y = y_train, family = "binomial",
                            alpha = 0, lambda = 1e-8,
                            intercept = TRUE, standardize = FALSE)
      prob <- as.numeric(stats::predict(fit, newx = X_sub_test, s = 1e-8,
                                        type = "response"))
    }
    return(as.numeric(prob > 0.5))
  }

  # multinomial
  X_refit <- X_sub_train
  X_test_refit <- X_sub_test
  if (ncol(X_refit) == 1) {
    X_refit      <- cbind(X_refit, 0)
    X_test_refit <- cbind(X_test_refit, 0)
  }
  fit <- glmnet::glmnet(x = X_refit, y = y_train, family = "multinomial",
                        type.multinomial = "grouped", alpha = 0,
                        nlambda = 50, intercept = TRUE, standardize = TRUE,
                        thresh = 1e-5, maxit = 1e5)
  lam_refit <- min(fit$lambda)
  pred <- stats::predict(fit, newx = X_test_refit, s = lam_refit, type = "class")
  as.character(if (is.matrix(pred)) pred[, 1] else pred)
}

---
title: "Best subset selection with combss"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Best subset selection with combss}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
set.seed(1)
```

`combss` solves the best subset selection problem in generalised linear
models by reformulating it as a continuous optimisation over the
hypercube $[0, 1]^p$ and applying a Frank-Wolfe homotopy algorithm. The
inner ridge problem is solved with `glmnet`. Three families are
supported:

- `family = "gaussian"` (alias `"linear"`) — linear regression
- `family = "binomial"` — binary logistic regression
- `family = "multinomial"` — multinomial logistic regression

```{r}
library(combss)
```

## Linear regression

A simulated dataset with $n = 300$ observations and $p = 30$ predictors,
of which only the first five are truly active.

```{r}
set.seed(102)
n <- 300; p <- 30
beta <- c(3, 2, 1.5, 1, 0.5, rep(0, p - 5))
x <- matrix(rnorm(n * p), n, p)
y <- as.numeric(x %*% beta + rnorm(n) * 0.5)

itr <- 1:200; iva <- 201:300
fit <- combss(x[itr, ], y[itr],
              x_val = x[iva, ], y_val = y[iva],
              family = "gaussian", q = 10)
fit
```

`fit$subset_list` contains the selected feature set for each `k = 1, ..., q`:

```{r}
fit$subset_list[1:8]
```

`fit$mse_path` gives the validation MSE at each `k`, and `fit$best_k`
the size with smallest MSE.

```{r}
plot(seq_along(fit$mse_path), fit$mse_path, type = "b",
     xlab = "k", ylab = "Validation MSE")
abline(v = fit$best_k, lty = 2)
```

## Binary logistic regression

```{r}
ybin <- as.numeric(plogis(x %*% beta) > 0.5)
fit_bin <- combss(x[itr, ], ybin[itr],
                  x_val = x[iva, ], y_val = ybin[iva],
                  family = "binomial", q = 10)
fit_bin
```

## Multinomial logistic regression

```{r}
ymulti <- cut(as.numeric(x %*% beta),
              breaks = c(-Inf, -1, 1, Inf),
              labels = c("a", "b", "c"))
fit_mn <- combss(x[itr, ], ymulti[itr],
                 x_val = x[iva, ], y_val = ymulti[iva],
                 family = "multinomial", q = 10)
fit_mn
```

## LOOCV ridge selection

For each candidate ridge penalty `lam_ridge`, `combss_cv()` runs COMBSS
and evaluates LOOCV error on the chosen subset.



```{r, eval = FALSE}
cv <- combss_cv(x, y, family = "gaussian", q = 6)
cv$best_lambda
```


## Predicting on new data

`predict()` refits on the chosen subset of the original training data
and predicts on `newx`.

```{r}
preds <- predict(fit, newx = x[iva, ],
                 x_train = x[itr, ], y_train = y[itr])
head(preds)
```

## References

- Moka, S., Liquet, B., Zhu, H. and Muller, S. (2024).
  COMBSS: best subset selection via continuous optimization.
  *Statistics and Computing*. \doi{10.1007/s11222-024-10387-8}
- Mathur, A., Liquet, B., Muller, S. and Moka, S. (2026).
  Parsimonious Subset Selection for Generalized Linear Models with
  Biomedical Applications. arXiv:2603.21952.

\name{predict.combss}
\alias{predict.combss}
\title{Predict from a \code{combss} fit.}
\usage{
\method{predict}{combss}(object, newx, x_train, y_train, k = NULL, ...)
}
\arguments{
\item{object}{A \code{combss} object.}

\item{newx}{Numeric matrix of new observations.}

\item{x_train, y_train}{Original training data used to refit on the chosen
subset.}

\item{k}{Subset size to use. Defaults to the best \code{k} if validation data
was used, else the largest \code{k} evaluated.}

\item{...}{Unused.}
}
\value{
Numeric vector of predictions for \code{newx}.
}
\description{
Refits the chosen subset on the original training data via OLS or ridge
(Gaussian / binomial / multinomial as appropriate) and predicts on \code{newx}.
Note: the training data is not stored on the fit object, so \code{newx} must be
supplied along with \code{x_train} and \code{y_train}.
}
\examples{
set.seed(1)
x <- matrix(rnorm(100 * 20), 100, 20)
y <- as.numeric(x[, 1] + x[, 2] + rnorm(100) * 0.1)
fit <- combss(x, y, family = "gaussian", q = 5)
newx <- matrix(rnorm(5 * 20), 5, 20)
predict(fit, newx = newx, x_train = x, y_train = y, k = 2)
}

\name{combss}
\alias{combss}
\title{Best subset selection for generalised linear models via continuous
optimisation (COMBSS).}
\usage{
combss(
  x,
  y,
  family = c("gaussian", "binomial", "multinomial"),
  q = NULL,
  x_val = NULL,
  y_val = NULL,
  lam_ridge = 0,
  Niter = 25,
  alpha = 0.01,
  scale = TRUE,
  mandatory = NULL,
  verbose = FALSE
)
}
\arguments{
\item{x}{Numeric design matrix \verb{(n, p)} without an intercept column.}

\item{y}{Response. Numeric for \code{family = "gaussian"}; \verb{\{0, 1\}} (or two-level
factor) for \code{family = "binomial"}; factor or integer in \verb{\{1, ..., C\}} for
\code{family = "multinomial"}.}

\item{family}{One of \code{"gaussian"} (alias: \code{"linear"}), \code{"binomial"},
\code{"multinomial"}.}

\item{q}{Maximum subset size. Defaults to \code{min(n, p)}.}

\item{x_val, y_val}{Optional validation data. When provided, the best \code{k}
is selected by validation metric and \code{coef_} is returned for the best
subset.}

\item{lam_ridge}{Ridge penalty for the inner solver. Default \code{0}.}

\item{Niter}{Number of homotopy iterations. Default \code{25}.}

\item{alpha}{Frank-Wolfe step size. Default \code{0.01}.}

\item{scale}{Column-normalise \code{x} before the algorithm. Default \code{TRUE}.}

\item{mandatory}{Optional integer vector of 1-indexed columns of \code{x} that
must be included in every model.}

\item{verbose}{Print progress. Default \code{FALSE}.}
}
\value{
An object of class \code{"combss"}, a list with elements:
\itemize{
\item \code{family}: one of \code{"gaussian"}, \code{"binomial"}, \code{"multinomial"}
\item \code{subset_list}: list of length up to \code{q}; element \code{k} is an integer
vector of selected column indices (1-indexed) for subset size \code{k}
\item \code{subset}: best subset (requires validation data; otherwise \code{NULL})
\item \code{mse} / \code{accuracy}: validation metric for the best subset
(requires validation data)
\item \code{lam_ridge}: ridge penalty used
\item \code{q}: maximum subset size evaluated
}
}
\description{
Reformulates the NP-hard discrete subset selection problem as a continuous
optimisation over the hypercube \verb{[0, 1]^p}, solved via a Frank-Wolfe
homotopy algorithm. The inner ridge problem is solved with \code{glmnet}. For
each subset size \verb{k = 1, ..., q}, COMBSS returns the selected feature set;
if validation data is supplied, the best \code{k} is chosen by validation MSE
(Gaussian) or accuracy (binomial / multinomial).
}
\examples{
set.seed(1)
n <- 200; p <- 30
beta <- c(3, 2, 1.5, 1, 0.5, rep(0, p - 5))
x <- matrix(rnorm(n * p), n, p)
y <- as.numeric(x \%*\% beta + rnorm(n) * 0.5)
fit <- combss(x, y, family = "gaussian", q = 10)
fit$subset_list[1:5]

}
\references{
Moka, S., Liquet, B., Zhu, H. and Muller, S. (2024).
COMBSS: best subset selection via continuous optimization.
\emph{Statistics and Computing}. \doi{10.1007/s11222-024-10387-8}

Mathur, A., Liquet, B., Muller, S. and Moka, S. (2026).
Parsimonious Subset Selection for Generalized Linear Models with
Biomedical Applications. \href{https://arxiv.org/abs/2603.21952}{arXiv:2603.21952}.
}

\name{coef.combss}
\alias{coef.combss}
\title{Extract the subset of selected feature indices.}
\usage{
\method{coef}{combss}(object, k = NULL, ...)
}
\arguments{
\item{object}{A \code{combss} object.}

\item{k}{Subset size to return. Defaults to the best \code{k} if validation
data was used, else the largest \code{k} evaluated.}

\item{...}{Unused.}
}
\value{
Integer vector of selected column indices (1-indexed).
}
\description{
Extract the subset of selected feature indices.
}
\examples{
set.seed(1)
x <- matrix(rnorm(100 * 20), 100, 20)
y <- as.numeric(x[, 1] + x[, 2] + rnorm(100) * 0.1)
fit <- combss(x, y, family = "gaussian", q = 5)
coef(fit, k = 2)
}

\docType{package}
\name{combss-package}
\alias{combss-package}
\title{combss: Best Subset Selection for Generalised Linear Models via Continuous Optimisation}
\description{
Best subset selection in linear, binary logistic, and multinomial logistic
regression via a Frank-Wolfe homotopy algorithm. The algorithm reformulates
the NP-hard discrete subset selection problem as a continuous optimisation
over the hypercube \verb{[0, 1]^p}, scaling to high-dimensional settings with
p >> n. Inner ridge solves use \pkg{glmnet}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/benoit-liquet/combss}
}

}
\author{
\strong{Maintainer}: Benoit Liquet \email{benoit.liquet@univ-pau.fr} (\href{https://orcid.org/0000-0002-8136-2294}{ORCID})

Authors:
\itemize{
  \item Anant Mathur
  \item Sarat Moka
}

}
\keyword{internal}

\name{combss_cv}
\alias{combss_cv}
\title{Select the ridge penalty for COMBSS by leave-one-out cross-validation.}
\usage{
combss_cv(
  x,
  y,
  family = c("gaussian", "binomial", "multinomial"),
  q,
  lambda_grid = NULL,
  Niter = 25,
  verbose = FALSE
)
}
\arguments{
\item{x}{Numeric design matrix \verb{(n, p)}.}

\item{y}{Response (numeric / binary / factor depending on \code{family}).}

\item{family}{One of \code{"gaussian"} (alias \code{"linear"}), \code{"binomial"},
\code{"multinomial"}.}

\item{q}{Maximum subset size.}

\item{lambda_grid}{Candidate ridge penalties. Defaults to
\code{c(0, 10^seq(-3, 1, length.out = 10))}.}

\item{Niter}{Number of homotopy iterations passed to \code{combss()}.}

\item{verbose}{Print progress.}
}
\value{
List with:
\itemize{
\item \code{best_lambda}: ridge penalty with best mean LOOCV metric across \code{k}
\item \code{best_lambda_per_k}: numeric vector, length \code{q}
\item \code{results}: data frame \verb{(lambda, k, loocv_metric)}
}
}
\description{
Mirrors \code{combss.cv.select_lambda} in the Python \code{combss} package. For each
candidate \code{lambda} and each subset size \verb{k = 1, ..., q}, runs COMBSS to
obtain a subset and evaluates LOOCV error of a refit. Returns the best
\code{lambda} overall and the best \code{lambda} per \code{k}.
}
